package jdk.internal.org.objectweb.asm.tree.analysis;

import jdk.internal.org.objectweb.asm.tree.AbstractInsnNode;
import jdk.internal.org.objectweb.asm.tree.LabelNode;
import jdk.internal.org.objectweb.asm.tree.IincInsnNode;
import jdk.internal.org.objectweb.asm.tree.VarInsnNode;
import jdk.internal.org.objectweb.asm.tree.TableSwitchInsnNode;
import jdk.internal.org.objectweb.asm.tree.LookupSwitchInsnNode;
import jdk.internal.org.objectweb.asm.Type;
import jdk.internal.org.objectweb.asm.tree.JumpInsnNode;
import java.util.HashMap;
import java.util.ArrayList;
import jdk.internal.org.objectweb.asm.tree.MethodNode;
import jdk.internal.org.objectweb.asm.tree.TryCatchBlockNode;
import java.util.List;
import jdk.internal.org.objectweb.asm.tree.InsnList;
import jdk.internal.org.objectweb.asm.Opcodes;

public class Analyzer<V extends Value> implements Opcodes
{
    private final Interpreter<V> interpreter;
    private int n;
    private InsnList insns;
    private List<TryCatchBlockNode>[] handlers;
    private Frame<V>[] frames;
    private Subroutine[] subroutines;
    private boolean[] queued;
    private int[] queue;
    private int top;
    
    public Analyzer(Interpreter<V> interpreter) {
        this.interpreter = interpreter;
    }
    
    public Frame<V>[] analyze(String internalName, MethodNode methodNode) throws AnalyzerException {
        if ((methodNode.access & 0x500) != 0x0) {
            return this.frames = (Frame[])new Frame[0];
        }
        this.n = methodNode.instructions.size();
        this.insns = methodNode.instructions;
        this.handlers = (List[])new List[this.n];
        this.frames = (Frame[])new Frame[this.n];
        this.subroutines = new Subroutine[this.n];
        this.queued = new boolean[this.n];
        this.queue = new int[this.n];
        this.top = 0;
        for (int i = 0; i < methodNode.tryCatchBlocks.size(); ++i) {
            TryCatchBlockNode tryCatchBlockNode = (TryCatchBlockNode)methodNode.tryCatchBlocks.get(i);
            int j;
            int index;
            List<TryCatchBlockNode> list;
            for (j = this.insns.indexOf(tryCatchBlockNode.start), index = this.insns.indexOf(tryCatchBlockNode.end), ; j < index; ++j) {
                list = this.handlers[j];
                if (list == null) {
                    list = new ArrayList();
                    this.handlers[j] = list;
                }
                list.add(tryCatchBlockNode);
            }
        }
        Subroutine o = new Subroutine(null, methodNode.maxLocals, null);
        ArrayList o2 = new ArrayList();
        HashMap o3 = new HashMap();
        this.findSubroutine(0, o, o2);
        while (!o2.isEmpty()) {
            JumpInsnNode jumpInsnNode = (JumpInsnNode)o2.remove(0);
            Subroutine subroutine = (Subroutine)o3.get(jumpInsnNode.label);
            if (subroutine == null) {
                Subroutine o4 = new Subroutine(jumpInsnNode.label, methodNode.maxLocals, jumpInsnNode);
                o3.put(jumpInsnNode.label, o4);
                this.findSubroutine(this.insns.indexOf(jumpInsnNode.label), o4, o2);
            }
            else {
                subroutine.callers.add(jumpInsnNode);
            }
        }
        for (int k = 0; k < this.n; ++k) {
            if (this.subroutines[k] != null && this.subroutines[k].start == null) {
                this.subroutines[k] = null;
            }
        }
        Frame frame = this.newFrame(methodNode.maxLocals, methodNode.maxStack);
        Frame<V> frame2 = this.newFrame(methodNode.maxLocals, methodNode.maxStack);
        frame.setReturn(this.interpreter.newValue(Type.getReturnType(methodNode.desc)));
        Type[] argumentTypes = Type.getArgumentTypes(methodNode.desc);
        int l = 0;
        if ((methodNode.access & 0x8) == 0x0) {
            frame.setLocal(l++, this.interpreter.newValue(Type.getObjectType(internalName)));
        }
        for (int n = 0; n < argumentTypes.length; ++n) {
            frame.setLocal(l++, this.interpreter.newValue(argumentTypes[n]));
            if (argumentTypes[n].getSize() == 2) {
                frame.setLocal(l++, this.interpreter.newValue(null));
            }
        }
        while (l < methodNode.maxLocals) {
            frame.setLocal(l++, this.interpreter.newValue(null));
        }
        this.merge(0, frame, null);
        this.init(internalName, methodNode);
        while (this.top > 0) {
            int[] queue = this.queue;
            int top = this.top - 1;
            this.top = top;
            int m = queue[top];
            Frame<V> frame3 = this.frames[m];
            Subroutine subroutine2 = this.subroutines[m];
            this.queued[m] = false;
            AbstractInsnNode value = null;
            try {
                value = methodNode.instructions.get(m);
                int opcode = value.getOpcode();
                int type = value.getType();
                if (type == 8 || type == 15 || type == 14) {
                    this.merge(m + 1, frame3, subroutine2);
                    this.newControlFlowEdge(m, m + 1);
                }
                else {
                    frame.init(frame3).execute(value, this.interpreter);
                    subroutine2 = ((subroutine2 == null) ? null : subroutine2.copy());
                    if ((value instanceof JumpInsnNode)) {
                        JumpInsnNode jumpInsnNode2 = (JumpInsnNode)value;
                        if (opcode != 167 && opcode != 168) {
                            this.merge(m + 1, frame, subroutine2);
                            this.newControlFlowEdge(m, m + 1);
                        }
                        int index7 = this.insns.indexOf(jumpInsnNode2.label);
                        if (opcode == 168) {
                            this.merge(index7, frame, new Subroutine(jumpInsnNode2.label, methodNode.maxLocals, jumpInsnNode2));
                        }
                        else {
                            this.merge(index7, frame, subroutine2);
                        }
                        this.newControlFlowEdge(m, index7);
                    }
                    else if ((value instanceof LookupSwitchInsnNode)) {
                        LookupSwitchInsnNode lookupSwitchInsnNode = (LookupSwitchInsnNode)value;
                        int index5 = this.insns.indexOf(lookupSwitchInsnNode.dflt);
                        this.merge(index5, frame, subroutine2);
                        this.newControlFlowEdge(m, index5);
                        for (int n4 = 0; n4 < lookupSwitchInsnNode.labels.size(); ++n4) {
                            int index6 = this.insns.indexOf((LabelNode)lookupSwitchInsnNode.labels.get(n4));
                            this.merge(index6, frame, subroutine2);
                            this.newControlFlowEdge(m, index6);
                        }
                    }
                    else if ((value instanceof TableSwitchInsnNode)) {
                        TableSwitchInsnNode tableSwitchInsnNode = (TableSwitchInsnNode)value;
                        int index3 = this.insns.indexOf(tableSwitchInsnNode.dflt);
                        this.merge(index3, frame, subroutine2);
                        this.newControlFlowEdge(m, index3);
                        for (int n3 = 0; n3 < tableSwitchInsnNode.labels.size(); ++n3) {
                            int index4 = this.insns.indexOf((LabelNode)tableSwitchInsnNode.labels.get(n3));
                            this.merge(index4, frame, subroutine2);
                            this.newControlFlowEdge(m, index4);
                        }
                    }
                    else if (opcode == 169) {
                        if (subroutine2 == null) {
                            throw new AnalyzerException(value, "RET instruction outside of a sub routine");
                        }
                        for (int n2 = 0; n2 < subroutine2.callers.size(); ++n2) {
                            int index2 = this.insns.indexOf((JumpInsnNode)subroutine2.callers.get(n2));
                            if (this.frames[index2] != null) {
                                this.merge(index2 + 1, this.frames[index2], frame, this.subroutines[index2], subroutine2.access);
                                this.newControlFlowEdge(m, index2 + 1);
                            }
                        }
                    }
                    else if (opcode != 191 && (opcode < 172 || opcode > 177)) {
                        if (subroutine2 != null) {
                            if ((value instanceof VarInsnNode)) {
                                int var = ((VarInsnNode)value).var;
                                subroutine2.access[var] = true;
                                if (opcode == 22 || opcode == 24 || opcode == 55 || opcode == 57) {
                                    subroutine2.access[var + 1] = true;
                                }
                            }
                            else if ((value instanceof IincInsnNode)) {
                                subroutine2.access[((IincInsnNode)value).var] = true;
                            }
                        }
                        this.merge(m + 1, frame, subroutine2);
                        this.newControlFlowEdge(m, m + 1);
                    }
                }
                List<Object> list2 = this.handlers[m];
                if (list2 != null) {
                    for (int n5 = 0; n5 < list2.size(); ++n5) {
                        TryCatchBlockNode tryCatchBlockNode2 = (TryCatchBlockNode)list2.get(n5);
                        Type type2;
                        if (tryCatchBlockNode2.type == null) {
                            type2 = Type.getObjectType("java/lang/Throwable");
                        }
                        else {
                            type2 = Type.getObjectType(tryCatchBlockNode2.type);
                        }
                        int index8 = this.insns.indexOf(tryCatchBlockNode2.handler);
                        if (this.newControlFlowExceptionEdge(m, tryCatchBlockNode2)) {
                            frame2.init(frame3);
                            frame2.clearStack();
                            frame2.push(this.interpreter.newValue(type2));
                            this.merge(index8, frame2, subroutine2);
                        }
                    }
                }
                else {
                    continue;
                }
            }
            catch (AnalyzerException ex) {
                throw new AnalyzerException(ex.node, "Error at instruction " + m + ": " + ex.getMessage(), ex);
            }
            catch (Exception ex2) {
                throw new AnalyzerException(value, "Error at instruction " + m + ": " + ex2.getMessage(), ex2);
            }
        }
        return this.frames;
    }
    
    private void findSubroutine(int index, Subroutine subroutine, List<AbstractInsnNode> list) throws AnalyzerException {
        while (index >= 0 && index < this.n) {
            if (this.subroutines[index] != null) {
                return;
            }
            this.subroutines[index] = subroutine.copy();
            AbstractInsnNode value = this.insns.get(index);
            if ((value instanceof JumpInsnNode)) {
                if (value.getOpcode() == 168) {
                    list.add(value);
                }
                else {
                    this.findSubroutine(this.insns.indexOf(((JumpInsnNode)value).label), subroutine, list);
                }
            }
            else if ((value instanceof TableSwitchInsnNode)) {
                TableSwitchInsnNode tableSwitchInsnNode = (TableSwitchInsnNode)value;
                this.findSubroutine(this.insns.indexOf(tableSwitchInsnNode.dflt), subroutine, list);
                for (int j = tableSwitchInsnNode.labels.size() - 1; j >= 0; --j) {
                    this.findSubroutine(this.insns.indexOf((LabelNode)tableSwitchInsnNode.labels.get(j)), subroutine, list);
                }
            }
            else if ((value instanceof LookupSwitchInsnNode)) {
                LookupSwitchInsnNode lookupSwitchInsnNode = (LookupSwitchInsnNode)value;
                this.findSubroutine(this.insns.indexOf(lookupSwitchInsnNode.dflt), subroutine, list);
                for (int i = lookupSwitchInsnNode.labels.size() - 1; i >= 0; --i) {
                    this.findSubroutine(this.insns.indexOf((LabelNode)lookupSwitchInsnNode.labels.get(i)), subroutine, list);
                }
            }
            List<Object> list2 = this.handlers[index];
            if (list2 != null) {
                for (int k = 0; k < list2.size(); ++k) {
                    this.findSubroutine(this.insns.indexOf(((TryCatchBlockNode)list2.get(k)).handler), subroutine, list);
                }
            }
            switch (value.getOpcode()) {
                case 167:
                case 169:
                case 170:
                case 171:
                case 172:
                case 173:
                case 174:
                case 175:
                case 176:
                case 177:
                case 191: {
                    return;
                }
                default: {
                    ++index;
                    continue;
                }
            }
        }
        throw new AnalyzerException(null, "Execution can fall off end of the code");
    }
    
    public Frame<V>[] getFrames() {
        return this.frames;
    }
    
    public List<TryCatchBlockNode> getHandlers(int n) {
        return this.handlers[n];
    }
    
    protected void init(String s, MethodNode methodNode) throws AnalyzerException {
    }
    
    protected Frame<V> newFrame(int n, int n2) {
        return new Frame(n, n2);
    }
    
    protected Frame<V> newFrame(Frame<? extends V> frame) {
        return new Frame(frame);
    }
    
    protected void newControlFlowEdge(int n, int n2) {
    }
    
    protected boolean newControlFlowExceptionEdge(int n, int n2) {
        return true;
    }
    
    protected boolean newControlFlowExceptionEdge(int n, TryCatchBlockNode tryCatchBlockNode) {
        return this.newControlFlowExceptionEdge(n, this.insns.indexOf(tryCatchBlockNode.handler));
    }
    
    private void merge(int n, Frame<V> frame, Subroutine subroutine) throws AnalyzerException {
        Frame<Object> frame2 = this.frames[n];
        Subroutine subroutine2 = this.subroutines[n];
        int merge;
        if (frame2 == null) {
            this.frames[n] = this.newFrame(frame);
            merge = true;
        }
        else {
            merge = frame2.merge(frame, this.interpreter);
        }
        if (subroutine2 == null) {
            if (subroutine != null) {
                this.subroutines[n] = subroutine.copy();
                merge = true;
            }
        }
        else if (subroutine != null) {
            merge |= subroutine2.merge(subroutine);
        }
        if (merge && !this.queued[n]) {
            this.queued[n] = true;
            this.queue[this.top++] = n;
        }
    }
    
    private void merge(int n, Frame<V> frame, Frame<V> frame2, Subroutine subroutine, boolean[] array) throws AnalyzerException {
        Frame<Object> frame3 = this.frames[n];
        Subroutine subroutine2 = this.subroutines[n];
        frame2.merge(frame, array);
        int merge;
        if (frame3 == null) {
            this.frames[n] = this.newFrame(frame2);
            merge = true;
        }
        else {
            merge = frame3.merge(frame2, this.interpreter);
        }
        if (subroutine2 != null && subroutine != null) {
            merge |= subroutine2.merge(subroutine);
        }
        if (merge && !this.queued[n]) {
            this.queued[n] = true;
            this.queue[this.top++] = n;
        }
    }
}
